/***********************************
 * Copyright (C) 2001  Jason Carter
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ***********************************
 * FILE: GENRES.H
 * NAME: ID3EDIT - ID3 tag v 1.1 editor
 * DESCRIPTION: Genres character array
 * AUTHOR: Jason Carter
 ***********************************
 *
 * Initially modified for use with MPlayer by Diego Biurrun on 2004-08-22
 * $Id: genres.h 31920 2010-08-04 12:20:01Z siretart $
 * detailed changelog at http://svn.mplayerhq.hu/mplayer/trunk/
 */

#ifndef MPLAYER_GENRES_H
#define MPLAYER_GENRES_H


const char * const genres[] = {
"Blues",					/* 0 */
"Classic Rock",				/* 1 */
"Country",					/* 2 */
"Dance",					/* 3 */
"Disco",					/* 4 */
"Funk",						/* 5 */
"Grunge",					/* 6 */
"Hip-Hop",					/* 7 */
"Jazz",						/* 8 */
"Metal",					/* 9 */
"New Age",					/* 10 */
"Oldies",					/* 11 */
"Other",					/* 12 */
"Pop",						/* 13 */
"R&B",						/* 14 */
"Rap",						/* 15 */
"Reggae",					/* 16 */
"Rock",						/* 17 */
"Techno",					/* 18 */
"Industrial",				/* 19 */
"Alternative",				/* 20 */
"Ska",						/* 21 */
"Death Metal",				/* 22 */
"Pranks",					/* 23 */
"Soundtrack",				/* 24 */
"Eurotechno",				/* 25 */
"Ambient",					/* 26 */
"Trip-Hop",					/* 27 */
"Vocal",					/* 28 */
"Jazz+Funk",				/* 29 */
"Fusion",					/* 30 */
"Trance",					/* 31 */
"Classical",				/* 32 */
"Instrumental",				/* 33 */
"Acid",						/* 34 */
"House",					/* 35 */
"Game",						/* 36 */
"Sound Clip",				/* 37 */
"Gospel",					/* 38 */
"Noise",					/* 39 */
"Alternative Rock",			/* 40 */
"Bass",						/* 41 */
"Soul",						/* 42 */
"Punk",						/* 43 */
"Space",					/* 44 */
"Meditative",				/* 45 */
"Instrumental Pop",			/* 46 */
"Instrumental Rock",		/* 47 */
"Ethnic",					/* 48 */
"Gothic",					/* 49 */
"Darkwave",					/* 50 */
"Techno-Industrial",		/* 51 */
"Electronic",				/* 52 */
"Pop-Folk",					/* 53 */
"Eurodance",				/* 54 */
"Dream",					/* 55 */
"Southern Rock",			/* 56 */
"Comedy",					/* 57 */
"Cult",						/* 58 */
"Gangsta",					/* 59 */
"Top 40",					/* 60 */
"Christian Rap",			/* 61 */
"Pop/Funk",					/* 62 */
"Jungle",					/* 63 */
"Native American",			/* 64 */
"Cabaret",					/* 65 */
"New Wave",					/* 66 */
"Psychadelic",			/* 67 */ /* typo taken from the id3v2spec*/
"Rave",						/* 68 */
"Show Tunes",				/* 69 */
"Trailer",					/* 70 */
"Lo-Fi",					/* 71 */
"Tribal",					/* 72 */
"Acid Punk",				/* 73 */
"Acid Jazz",				/* 74 */
"Polka",					/* 75 */
"Retro",					/* 76 */
"Musical",					/* 77 */
"Rock & Roll",				/* 78 */
"Hard Rock",				/* 79 */
"Folk",						/* 80 */
"Folk/Rock",				/* 81 */
"National Folk",			/* 82 */
"Swing",					/* 83 */
"Fast-Fusion",				/* 84 */
"Bebop",					/* 85 */
"Latin",					/* 86 */
"Revival",					/* 87 */
"Celtic",					/* 88 */
"Bluegrass",				/* 89 */
"Avantgarde",				/* 90 */
"Gothic Rock",				/* 91 */
"Progressive Rock",			/* 92 */
"Psychedelic Rock",			/* 93 */
"Symphonic Rock",			/* 94 */
"Slow Rock",				/* 95 */
"Big Band",					/* 96 */
"Chorus",					/* 97 */
"Easy Listening",			/* 98 */
"Acoustic",					/* 99 */
"Humour",					/* 100 */
"Speech",					/* 101 */
"Chanson",					/* 102 */
"Opera",					/* 103 */
"Chamber Music",			/* 104 */
"Sonata",					/* 105 */
"Symphony",					/* 106 */
"Booty Bass",				/* 107 */
"Primus",					/* 108 */
"Porn Groove",				/* 109 */
"Satire",					/* 110 */
"Slow Jam",					/* 111 */
"Club",						/* 112 */
"Tango",					/* 113 */
"Samba",					/* 114 */
"Folklore",					/* 115 */
"Ballad",					/* 116 */
"Power Ballad",				/* 117 */
"Rhytmic Soul",				/* 118 */
"Freestyle",				/* 119 */
"Duet",						/* 120 */
"Punk Rock",				/* 121 */
"Drum Solo",				/* 122 */
"Acapella",					/* 123 */
"Euro-House",				/* 124 */
"Dance Hall",				/* 125 */
"Goa",						/* 126 */
"Drum & Bass",				/* 127 */
"Club-House",				/* 128 */
"Hardcore",					/* 129 */
"Terror",					/* 130 */
"Indie",					/* 131 */
"BritPop",					/* 132 */
"Negerpunk",				/* 133 */
"Polsk Punk",				/* 134 */
"Beat",						/* 135 */
"Christian Gangsta Rap",	/* 136 */
"Heavy Metal",				/* 137 */
"Black Metal",				/* 138 */
"Crossover",				/* 139 */
"Contemporary Christian",	/* 140 */
"Christian Rock",			/* 141 */
"Merengue",					/* 142 */
"Salsa",					/* 143 */
"Thrash Metal",					/* 144 */
"Anime",					/* 145 */
"Jpop",					/* 146 */
"Synthpop",					/* 147 */
"Unknown",					/* 148 */
"Unknown",					/* 149 */
"Unknown",					/* 150 */
"Unknown",					/* 151 */
"Unknown",					/* 152 */
"Unknown",					/* 153 */
"Unknown",					/* 154 */
"Unknown",					/* 155 */
"Unknown",					/* 156 */
"Unknown",					/* 157 */
"Unknown",					/* 158 */
"Unknown",					/* 159 */
"Unknown",					/* 160 */
"Unknown",					/* 161 */
"Unknown",					/* 162 */
"Unknown",					/* 163 */
"Unknown",					/* 164 */
"Unknown",					/* 165 */
"Unknown",					/* 166 */
"Unknown",					/* 167 */
"Unknown",					/* 168 */
"Unknown",					/* 169 */
"Unknown",					/* 170 */
"Unknown",					/* 171 */
"Unknown",					/* 172 */
"Unknown",					/* 173 */
"Unknown",					/* 174 */
"Unknown",					/* 175 */
"Unknown",					/* 176 */
"Unknown",					/* 177 */
"Unknown",					/* 178 */
"Unknown",					/* 179 */
"Unknown",					/* 180 */
"Unknown",					/* 181 */
"Unknown",					/* 182 */
"Unknown",					/* 183 */
"Unknown",					/* 184 */
"Unknown",					/* 185 */
"Unknown",					/* 186 */
"Unknown",					/* 187 */
"Unknown",					/* 188 */
"Unknown",					/* 189 */
"Unknown",					/* 190 */
"Unknown",					/* 191 */
"Unknown",					/* 192 */
"Unknown",					/* 193 */
"Unknown",					/* 194 */
"Unknown",					/* 195 */
"Unknown",					/* 196 */
"Unknown",					/* 197 */
"Unknown",					/* 198 */
"Unknown",					/* 199 */
"Unknown",					/* 200 */
"Unknown",					/* 201 */
"Unknown",					/* 202 */
"Unknown",					/* 203 */
"Unknown",					/* 204 */
"Unknown",					/* 205 */
"Unknown",					/* 206 */
"Unknown",					/* 207 */
"Unknown",					/* 208 */
"Unknown",					/* 209 */
"Unknown",					/* 210 */
"Unknown",					/* 211 */
"Unknown",					/* 212 */
"Unknown",					/* 213 */
"Unknown",					/* 214 */
"Unknown",					/* 215 */
"Unknown",					/* 216 */
"Unknown",					/* 217 */
"Unknown",					/* 218 */
"Unknown",					/* 219 */
"Unknown",					/* 220 */
"Unknown",					/* 221 */
"Unknown",					/* 222 */
"Unknown",					/* 223 */
"Unknown",					/* 224 */
"Unknown",					/* 225 */
"Unknown",					/* 226 */
"Unknown",					/* 227 */
"Unknown",					/* 228 */
"Unknown",					/* 229 */
"Unknown",					/* 230 */
"Unknown",					/* 231 */
"Unknown",					/* 232 */
"Unknown",					/* 233 */
"Unknown",					/* 234 */
"Unknown",					/* 235 */
"Unknown",					/* 236 */
"Unknown",					/* 237 */
"Unknown",					/* 238 */
"Unknown",					/* 239 */
"Unknown",					/* 240 */
"Unknown",					/* 241 */
"Unknown",					/* 242 */
"Unknown",					/* 243 */
"Unknown",					/* 244 */
"Unknown",					/* 245 */
"Unknown",					/* 246 */
"Unknown",					/* 247 */
"Unknown",					/* 248 */
"Unknown",					/* 249 */
"Unknown",					/* 250 */
"Unknown",					/* 251 */
"Unknown",					/* 252 */
"Unknown",					/* 253 */
"Unknown",					/* 254 */
"Unknown",					/* 255 */
};

#endif /* MPLAYER_GENRES_H */
