/*
 * Copyright (C) 2002 the xine project
 *
 * This file is part of xine, a free video player.
 *
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * $Id: pnm.h 29305 2009-05-13 02:58:57Z diego $
 *
 * pnm util functions header by joschka
 */

#ifndef MPLAYER_PNM_H
#define MPLAYER_PNM_H

#include <inttypes.h>
/*#include "xine_internal.h" */

typedef struct pnm_s pnm_t;

//pnm_t*   pnm_connect (int fd,char *url);

//int      pnm_read (pnm_t *this, char *data, int len);

#endif /* MPLAYER_PNM_H */
