LIBAVCODEC_$MAJOR {
        global: *;
        local:
                ff_*_bsf;
                ff_*_decoder;
                ff_*_encoder;
                ff_*_hwaccel;
                ff_*_parser;
};
